/*
 * This is a simple batch-mode recording program.
 *
 * The basic recording functionality was taken from wavnorm 0.4.  This was
 * written by "Duncan" (duncan@zog.net.au).
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <limits.h>

#include <math.h>
#include <signal.h>

#include "wavutils.h"
#include "text_display.h"
#include "version.h"

/* ======================================================================== */

/** this is for recording to CD so this is the only setting I want **/
#define SAMPLES_PER_SECOND 44100
#define BITS_PER_SAMPLE    16
#define NUMBER_OF_CHANNELS 2

#define BYTES_PER_SAMPLE   (BITS_PER_SAMPLE/8)
#define BYTES_PER_SECOND   (SAMPLES_PER_SECOND*NUMBER_OF_CHANNELS*BYTES_PER_SAMPLE)

#define DEFAULT_DSP_DEVNAME "/dev/dsp"
#define DEFAULT_TIMEOUT     2

/* VU meter timings in seconds.  For best results MAX_HOLDTIME should be a */
/* multiple of VU_REFRESH. */
#define MAX_HOLDTIME        0.20
#define VU_REFRESH          0.05

char *dsp_devname = NULL;
size_t dsp_blocksize = 0;
int user_exit_req = 0;
int null_recording = 0;
float threshold,timeout = -1.0;
char *filename = NULL;

/* ======================================================================== */

void sig_int_handler(int signum) {
/*
 * Purpose: handles a SIGINT by signaling to the program to stop.
 *
 */
  user_exit_req = 1;
  /* Don't reinstall signal handler during testing */
}
/* ======================================================================== */
/* Routines to convert to/from dB.  Assumes incoming data ranges from
 * -32768 to +32767.
 */

float to_dB(float source, int *sign) {
  if (sign != NULL)
    *sign = (source > 0)?1:-1;
  if (source != 0)
    return(20.0 * log10(fabs(source)/32767.0));
  else
    return(20.0 * log10(0.5/32767.0));
}
float from_dB(float source) {
  if (source >= -96)
    return(pow(10,source/20.0)*32767.0);
  else
    return(0);
}
/* ======================================================================== */

int open_dsp_read(char *dsp_filename) {
/*
 * Purpose: opens up /dev/dsp to read audio data.
 *
 * Notes: Presently forces 16 bit stereo at 44100Hz.
 *        Returns the fd to the dsp and sets the block size in blk_size.
 */
int dspfd;
int status;
int arg;

  /** open the sound device and set it up **/
  dspfd = open(dsp_filename,O_RDONLY);
  if (dspfd < 0) {
    printf("Error: open(%s) %s\n", dsp_filename,strerror(errno));
    return(-1);
  }
  arg = BITS_PER_SAMPLE;
  status = ioctl(dspfd, SOUND_PCM_WRITE_BITS, &arg);

  if (status == -1) {
    perror("SOUND_PCM_WRITE_BITS ioctl failed");
    return(-1);
  }
  if (arg != BITS_PER_SAMPLE) {
    perror("unable to set sample size");
    return(-1);
  }

  arg = NUMBER_OF_CHANNELS;  /* mono or stereo */
  status = ioctl(dspfd, SOUND_PCM_WRITE_CHANNELS, &arg);
  if (status == -1) {
    perror("SOUND_PCM_WRITE_CHANNELS ioctl failed");
    return(-1);
  }
  if (arg != NUMBER_OF_CHANNELS) {
    perror("unable to set number of channels");
    return(-1);
  }

  arg = SAMPLES_PER_SECOND;      /* sampling rate */
  status = ioctl(dspfd, SOUND_PCM_WRITE_RATE, &arg);
  if (status == -1) {
    perror("SOUND_PCM_WRITE_RATE ioctl failed");
    return(-1);
  }
  status = ioctl(dspfd, SNDCTL_DSP_GETBLKSIZE,&arg);
  /* printf("status %d, blksize %d\n",status,arg); */
  dsp_blocksize = arg;

  /* 100ms delay to avoid the blips when opening the dsp device */
  usleep(100*1000);

  printf("Input device is %s\n",dsp_filename);
  printf("Sampling %d channel(s)\n",NUMBER_OF_CHANNELS);
  printf("%d Hz sampling frequency, %d bits per sample\n",
    SAMPLES_PER_SECOND,BITS_PER_SAMPLE);
  printf("Using block size of %d bytes\n",dsp_blocksize);
  if (timeout == -1) 
    printf("Recording until <ctrl>-C is pressed\n");
  printf("\n");
  return(dspfd);
}
/* ======================================================================== */

void make_length_str(int n_samples, char *str) {
/*
 * Purpose: constructs a string describing the length of the current sample.
 *          "str" must point to a string buffer with space for at least 20 
 *          characters.
 *
 * Notes: assumes stereo recording with a sampling frequency of 44100Hz.
 */
int sec100s = (int)(n_samples / (SAMPLES_PER_SECOND/100.0));
int size = NUMBER_OF_CHANNELS * n_samples * BYTES_PER_SAMPLE;
char sizebuf[10];

  if (size < 9999)
    sprintf(sizebuf,"%4dB",size);
  else
  if (size < 9999*1024)
    sprintf(sizebuf,"%4dkB",size/1024);
  else
  if (size < 99.9*1024*1024)
    sprintf(sizebuf,"%4.1fMB",size/(1024*1024.0));
  else
  if (size < 2047*1024*1024)
    sprintf(sizebuf,"%4dMB",size/(1024*1024));
  else
    sprintf(sizebuf,"!!!!MB");

  sprintf(str,"%02d:%02d.%02d (%s)",sec100s/(100*60),(sec100s/100)%60,
    sec100s % 100,sizebuf);
}
/* ======================================================================== */

int record_wav_file(int dspfd, char *filename) {
/*
 * Purpose: records into a single given file.
 *
 */
char *wavbuf;
int wavfd;
size_t curr_blocksize;
int running = 1;
int recording = (timeout==-1);
float below_age = 0;
size_t bytes_written = 0;
char length_str[20];
         
WAVE_HEADER wav_header, *wav_header_on_disk;
short max[2], ave[2];
int acc_max[2], acc_ave[2];
int vu_count, i;
int max_age[2];

  if ((wavbuf = malloc(dsp_blocksize)) == NULL) {
    printf("Error: malloc() %s\n",strerror(errno));
    return(0);
  }

  create_wav_header(&wav_header,
    NUMBER_OF_CHANNELS,SAMPLES_PER_SECOND,BITS_PER_SAMPLE);
  wavfd = open(filename,O_RDWR | O_CREAT | O_TRUNC,0600);
  if (wavfd == -1) {
    printf("Error: open() %s\n",strerror(errno));
    free(wavbuf);
    return(0);
  }
  if (write(wavfd,(void *)&wav_header,sizeof(wav_header)) != sizeof(wav_header)) {
    printf("Error: write() %s\n",strerror(errno));
    close(wavfd);
    free(wavbuf);
    return(0);
  }
  if (timeout > 0) {
    printf("Ready to record %s\n\n",filename);
  }
  text_vu_headers(); 
  vu_count = 0;
  for (i=0;i<2;i++) {
    max_age[i] = 0;
    acc_ave[i] = acc_max[i] = 0;
  }
  while (running) {
    /* A short read could be due to a signal interrupt */
    curr_blocksize = read(dspfd,wavbuf,dsp_blocksize);
    if (curr_blocksize!=dsp_blocksize && errno!=0) {
      printf("read(dsp) %s",strerror(errno));
      return(0);
    }
    get_stereo_peak_values((short *)wavbuf,(short *)(wavbuf + curr_blocksize),
      &max[0], &max[1]);
    get_stereo_ave_values((short *)wavbuf,(short *)(wavbuf + curr_blocksize),
      &ave[0], &ave[1]);
    if (timeout>0) {
      recording = (recording || to_dB(max[0],NULL)>threshold ||
        to_dB(max[1],NULL)>threshold);
      if (recording) {
        if (to_dB(max[0],NULL)<threshold && to_dB(max[1],NULL)<threshold)
          below_age += curr_blocksize;
        else
          below_age = 0;
        running = (below_age/BYTES_PER_SECOND < timeout);
      }
    }
    if (!null_recording && recording) {
      if (write(wavfd,wavbuf,curr_blocksize) != curr_blocksize) {
        printf("write(%s) %s",filename,strerror(errno));
        return(0);
      }
      bytes_written += curr_blocksize;
    }

    for (i=0;i<2;i++) {
      if (max[i]>acc_max[i] || (float)max_age[i]/BYTES_PER_SECOND > MAX_HOLDTIME) {
        acc_max[i] = max[i];
        max_age[i] = 0;
      } else {
        max_age[i] += curr_blocksize;
      }
      acc_ave[i] += ave[i];
    }
    vu_count++;
    if ((float)vu_count*dsp_blocksize/BYTES_PER_SECOND > VU_REFRESH) {
      text_vu_values(to_dB(acc_max[0],NULL),to_dB(acc_max[1],NULL),
                     to_dB((float)acc_ave[0]/vu_count,NULL),
                     to_dB((float)acc_ave[1]/vu_count,NULL));
      vu_count = 0;
      for (i=0;i<2;i++)
        acc_ave[i] = 0;

      if (recording) {
        make_length_str(bytes_written/(BYTES_PER_SAMPLE*NUMBER_OF_CHANNELS),
          length_str);
        printf(length_str);
      } else
        printf("waiting...");
      fflush(stdout);
    }
    if (user_exit_req) {
      running = 0;
    }
  }
  printf("\n\n");

  /* Set the wav file header to have the right length */
  if (bytes_written == 0) {
    printf("Recording of %s finished with no data written: file erased.\n\n",filename);
    remove(filename);
  } else {
    wav_header_on_disk = (WAVE_HEADER *)mmap(NULL,sizeof(WAVE_HEADER),
      PROT_READ | PROT_WRITE,MAP_SHARED,wavfd,0);
    if (wav_header_on_disk == MAP_FAILED) {
      printf("Warning: mmap() %s\n",strerror(errno));
      printf("File %s will be OK but you may have to fix the length\n",filename);
    } else {
      wav_header_on_disk->nDataBytes = bytes_written;
      munmap(wav_header_on_disk,sizeof(WAVE_HEADER));
      make_length_str(bytes_written/(BYTES_PER_SAMPLE*NUMBER_OF_CHANNELS),
        length_str);
      printf("Recording of %s complete: final length %s\n",filename,length_str);
    }
  }
  close(wavfd);               

  free(wavbuf);
  return(1);
}
/* ======================================================================== */

void show_usage() {
/*
 * Purpose: gives the user some help on the program's functionality.
 *
 */

  fprintf(stderr,"batchrec version " VERSION "\n");
  fprintf(stderr,"(c) 2001 Jonathan Woithe (jwoithe@physics.adelaide.edu.au)\n"
                 "Released under the terms of the GNU General Public License.  No warranty is\n"
                 "given or implied, not even for fitness of purpose; see source for full details.\n\n"
                 "usage: batchrec [options] <filename>\n"
                 "Available options:\n"
                 " -d device           Use <device> as audio source\n"
                 " -h                  Display this help and exit\n"
                 " -n                  Null recording: do everything except write data to file\n"
                 " -t thres[,timeout]  Set threshold and timeout for multiple recording mode\n"
                 "\n");
  fprintf(stderr,"In normal mode, recording is made into the supplied filename until <ctrl>-C\n"
                 "is pressed.\n\n");
  fprintf(stderr,"In multiple recording mode, recording starts when the input exceeds the given\n"
                 "threshold (in dB) and continues until the signal stays below the threshold\n"
                 "for a period of time equal to the timeout.  The process repeats until <ctrl>-C\n"
                 "is pressed.  In this mode, the output files are generated by appending the\n"
                 "relevant recording number to the supplied filename.  If the supplied filename\n"
                 "ends in a valid number, that number (zero padded to 4 digits) is used as the\n"
                 "initial recording number.  The default timeout value is 2 seconds.\n");
}
/* ======================================================================== */

int parse_args(int argc, char *argv[]) {
/*
 * Purpose: parses the command line parameters.  Returns 1 on success.
 *
 */
int c;
int error = 0;
  while (!error && (c = getopt(argc, argv, "d:hnt:"))!=EOF) {
    switch(c) {
      case 'd':
        dsp_devname = strdup(optarg);
        break;
      case 'h':
        show_usage();
        return(0);
        break;
      case 'n':
        null_recording = 1;
        break;
      case 't': {
        char sep;
        int num = sscanf(optarg,"%g%c%g",&threshold,&sep,&timeout);
        if (num == 1)
          timeout = DEFAULT_TIMEOUT;
        else
        if (num==2 || sep!=',') {
          error = 1;
          fprintf(stderr,"batch_rec: malformed `-t' parameter\n");
        } else
        if (threshold>0 || timeout<=0) {
          error = 1;
          fprintf(stderr,"batch_rec: `-t' parameter values invalid\n");
        }
        break;
      }
      default:
        error = 1;
    }
  }
  if (!error && dsp_devname==NULL)
    dsp_devname = strdup(DEFAULT_DSP_DEVNAME);

  if (!error && (optind + 1)==argc) {
    filename = strdup(argv[optind]);
  } else {
    error = 1;
    fprintf(stderr,"batch_rec: no output filename specified\n");
  }

  return(!error);
}
/* ======================================================================== */

void cleanup() {
/*
 * Purpose: Performs final cleanup duties prior to exit.
 *
 */
  if (dsp_devname != NULL)
    free(dsp_devname);
  if (filename != NULL)
    free(filename);
}
/* ======================================================================== */

char *next_filename(char *name,int next) {
/*
 * Purpose: generates a sequential filename for use in multiple recording
 *          mode.  If "next" is true, the new filename will be the next number
 *          along (if the filename contains a number), or 0 if no number is
 *          present.  If "next" is false, the filename provided is simply
 *          made compatible with the sequential naming convention.
 */
char *baseend,*ext;
char *res;
int i = 0;
int num = 0;

  if ((baseend = strchr(name,'.')) == NULL) {
    ext = baseend = name + strlen(name);
  } else {
    ext = baseend+1;
    *baseend = '\0';
  }
  while (i<4 && *(baseend-1)>='0' && *(baseend-1)<='9') {
    i++;
    baseend--;
  }
  /* A number exists, so identify it */
  if (i>0) {
    sscanf(baseend,"%d",&num);
    if (next)
      num++;
    *baseend = '\0';
  }
  res = malloc(strlen(name)+4+1+strlen(ext)+1);
  sprintf(res,"%s%04d%s%s",name,num,strlen(ext)>0?".":"",ext);
  free(name);
  return(res);
}
/* ======================================================================== */

int main(int argc,char *argv[]) {
int dspfd,c;

  signal(SIGINT,sig_int_handler);
  if ((c = parse_args(argc,argv))) {

    /* Open the dsp device */
    dspfd = open_dsp_read(dsp_devname);
    if (dspfd == -1) {
      fprintf(stderr,"Error getting %s ready for recording\n",dsp_devname);
    } else {
      if (timeout > 0)
        filename = next_filename(filename,0);
      while (!user_exit_req) {
        record_wav_file(dspfd,filename);
        if (timeout > 0)
          filename = next_filename(filename,1);
      }
      close(dspfd);
    }
  }
  cleanup();
  return(!c);
}
